let utilities = require('../../system/utilities'),
  commons = require('happn-commons'),
  async = commons.async,
  _ = commons._;
module.exports = Rest;

/**
 * Rest component, exposes the exchange over a lightweight REST service
 * @constructor
 */
function Rest() {
  this.jsonBody = require('body/json');
}

// does a happn login, returns a session token that is usable for subsequent operations
Rest.prototype.login = function ($happn, req, res) {
  this.__parseBody(req, res, $happn, (body) => {
    let params = body;

    if (params.parameters) params = params.parameters;

    if (!params.username) {
      return this.__respond(
        $happn,
        'Failure parsing request body',
        null,
        new Error('no username'),
        res,
        401
      );
    }

    if (!params.password && !params.token) {
      return this.__respond(
        $happn,
        'Failure parsing request body',
        null,
        new Error('no password'),
        res,
        401
      );
    }

    this.__securityService.login(
      { username: params.username, password: params.password, token: params.token },
      undefined,
      req,
      (e, session) => {
        if (e) {
          if (
            e.toString().toLowerCase().includes('invalid credentials') ||
            [401, 403].includes(e.code)
          ) {
            return this.__respond($happn, 'Failure logging in', null, e, res, 401);
          }
          return this.__respond($happn, 'Failure logging in', null, e, res);
        }
        this.__respond($happn, 'Logged in ok', { token: session.token }, null, res);
      }
    );
  });
};

// attached to the mesh middleware, takes calls for a description of the API
Rest.prototype.describe = function ($happn, _req, res, $origin) {
  let description = utilities.clone(this.__exchangeDescription);

  if (!$origin || $origin.username === '_ADMIN') {
    return this.__respond(
      $happn,
      $happn._mesh.description.name + ' description',
      description.callMenu,
      null,
      res
    );
  }

  async.eachSeries(
    Object.keys(description.callMenu),
    (methodURI, methodURICB) => {
      this.__authorizeMethod($happn, $origin, methodURI, (e, authorized) => {
        if (e) return methodURICB(e);
        if (!authorized) delete description.callMenu[methodURI];
        methodURICB();
      });
    },
    (e) => {
      if (e) return this.__respond($happn, 'call failed', null, e, res);
      this.__respond(
        $happn,
        $happn._mesh.description.name + ' description',
        description.callMenu,
        null,
        res
      );
    }
  );
};

Rest.prototype.__respond = function ($happn, message, data, error, res, code) {
  const header = {
    'Content-Type': 'application/json',
  };

  const response = {
    message,
    data: typeof data === 'undefined' ? null : data,
    error: null,
  };

  if (error) {
    const plainError = utilities.plainError(error, false);
    if (!code) code = 500;
    response.error = plainError;
    $happn.log.warn(
      `rpc request failure: ${error.message ? error.message : JSON.stringify(plainError)}`
    );
  } else {
    if (!code) code = 200;
  }

  res.writeHead(code, header);
  res.end(JSON.stringify(response));
};

Rest.prototype.__authorizeMethod = function ($happn, $origin, methodURI, callback) {
  const requestPath = `/_exchange/requests/${
    $happn._mesh.config.domain || $happn._mesh.config.name
  }/${utilities.removeLeading('/', methodURI)}`;

  this.__securityService.authorize($origin, requestPath, 'set', function (e, authorized, reason) {
    callback(e, authorized, reason);
  });
};

Rest.prototype.__validateCredentialsGetOrigin = function (
  $happn,
  $origin,
  res,
  methodURI,
  authorizeAs,
  successful
) {
  if (!$happn._mesh.config.happn.secure) {
    return successful();
  }

  if (!$origin) {
    return this.__respond(
      $happn,
      'Bad origin',
      null,
      new Error('origin of call unknown'),
      res,
      403
    );
  }
  this.__getAuthorizedOrigin(
    $happn,
    $origin,
    res,
    methodURI,
    authorizeAs,
    (e, authorizedOrigin) => {
      // this will not get hit on a 403
      if (e) {
        if (e.message === 'origin does not belong to the delegate group') {
          return this.__respond($happn, 'Authorization failed', null, e, res, 403);
        }
        this.__respond($happn, 'Authorization failed due to system error', null, e, res, 500);
        return $happn.log.warn(`authorization system error: ${e.message}`);
      }
      successful(authorizedOrigin);
    }
  );
};

Rest.prototype.__getAuthorizedOrigin = function (
  $happn,
  $origin,
  res,
  methodURI,
  authorizeAs,
  callback
) {
  if (authorizeAs == null) {
    // backward compatible, if the incoming request is not being delegated, we authorize on the edge
    // then return the _ADMIN session so that method request further down the stack automatically work
    return this.__authorizeMethod($happn, $origin, methodURI, (e, authorized, reason) => {
      if (e) {
        return callback(e);
      }
      if (authorized === false) {
        // return 403 error
        if (!reason) reason = 'Authorization failed';
        return this.__respond($happn, reason, null, new Error('Access denied'), res, 403);
      }
      callback(null, Object.assign({}, $origin, { edgeAuthorized: true }));
    });
  }
  if ($origin.username === '_ADMIN') {
    return callback(null, Object.assign({}, $origin, { username: authorizeAs }));
  }
  let callbackWasCalled = false;
  this.__securityService.users
    .userBelongsToGroups($origin.username, ['_MESH_DELEGATE'])
    .then((belongs) => {
      callbackWasCalled = true;
      if (!belongs) {
        return callback(new Error('origin does not belong to the delegate group'));
      }
      return callback(null, Object.assign({}, $origin, { username: authorizeAs }));
    })
    .catch((e) => {
      if (callbackWasCalled) {
        return $happn.log.warn(
          `failure after getAuthorizedOrigin, and callback already called: origin ${$origin.username} authorizing as ${authorizeAs}}`
        );
      }
      return callback(e);
    });
};

Rest.prototype.__parseBody = function (req, res, $happn, callback) {
  try {
    if (req.method === 'GET') {
      if (req.uri.query && req.uri.query.encoded_parameters)
        return callback(JSON.parse(decodeURIComponent(req.uri.query.encoded_parameters)));
      return callback({ parameters: req.uri.query });
    }
    if (req.body) return callback(req.body);
    this.jsonBody(req, res, (e, body) => {
      if (e) return this.__respond($happn, 'Failure parsing request body', null, e, res, 400);
      callback(body);
    });
  } catch (e) {
    return this.__respond($happn, 'Failure parsing request body', null, e, res, 400);
  }
};

Rest.prototype.__processRequest = function (req, res, body, callPath, $happn, $origin) {
  process.nextTick(() => {
    let component;
    let method;
    let meshDescription;
    let componentDescription;
    let methodDescription;

    let methodName = callPath.pop();
    let componentName = callPath.pop();
    let meshName = callPath.pop();

    let mesh = $happn.as(
      $origin?.username,
      componentName,
      methodName,
      0,
      $origin?.edgeAuthorized
    ).exchange;

    if (componentName === 'security') {
      return this.__respond(
        $happn,
        'Access denied',
        null,
        new Error('attempt to access security component over rest'),
        res,
        403
      );
    }

    if (meshName != null) {
      if (meshName !== $happn._mesh.config.name) {
        return this.__respond(
          $happn,
          'Access denied',
          null,
          new Error('attempt to access remote mesh: ' + meshName),
          res,
          403
        );
      }
      mesh = $happn.exchange[meshName];
    }
    if (!mesh[componentName]) {
      return this.__respond(
        $happn,
        'Failure parsing request body',
        null,
        new Error('component ' + componentName + ' does not exist on mesh'),
        res,
        404
      );
    }

    component = mesh[componentName];

    if (!component[methodName]) {
      $happn.log.warn(
        `unable to find method: ${componentName}.${methodName} on api id: ${$happn.id}`
      );
      return this.__respond(
        $happn,
        'Failure parsing request body',
        null,
        new Error(`method ${methodName} does not exist on component ${componentName}`),
        res,
        404
      );
    }
    method = component[methodName];

    meshDescription = this.__exchangeDescription;
    componentDescription = meshDescription.components[componentName];
    methodDescription = componentDescription.methods[methodName];

    const args = this.__mapMethodArguments(req, res, methodDescription, body, $happn, $origin);
    method.apply(Object.assign({}, method, { $origin }), args);
  });
};

Rest.prototype.__mapMethodArguments = function (
  req,
  res,
  methodDescription,
  body,
  $happn,
  $origin
) {
  const __callback = (e, response) => {
    if (e) {
      if (e.message === 'unauthorized') {
        return this.__respond($happn, e.reason, null, new Error('Access denied'), res, 403);
      }
      //TODO: handle malformed errors as well
      return this.__respond($happn, 'Call failed', null, e, res, 500);
    }
    this.__respond($happn, 'Call successful', response, null, res);
  };

  // if we are passing in an array of arguments, they have already been intentionally mapped by the user
  if (Array.isArray(body)) {
    return body.concat([__callback]);
  }

  // if body.parameters is an array of arguments, they have already been intentionally mapped by the user
  if (Array.isArray(body.parameters)) {
    return body.parameters.concat([__callback]);
  }

  let callbackFound = false;
  if (!body.parameters) body.parameters = {};

  const args = methodDescription.parameters.map((parameter) => {
    if (parameter.name === 'callback') {
      callbackFound = true;
      return __callback;
    }

    if (parameter.name === '$restParams') {
      if (!body.parameters) return {};
      return _.omit(body.parameters, ['happn_token', 'username', 'password']);
    }

    if (parameter.name === '$userSession') {
      return $origin;
    }

    if (parameter.name.indexOf('$req_') === 0) {
      let key = parameter.name.split('$req_').pop();
      return req[key];
    }

    if (body.parameters[parameter.name] == null) {
      return null;
    }

    return body.parameters[parameter.name];
  });

  //add the callback handler
  if (!callbackFound) args.push(__callback);
  return args;
};

// attached to the mesh middleware, takes in the request body and attempts to execute an exchange method based on the request parameters
Rest.prototype.handleRequest = function (req, res, $happn, $origin) {
  try {
    const methodURI = utilities.removeLeading('/', utilities.getRelativePath(req.url));
    const callPath = methodURI.split('/');
    //ensure we don't have a leading /
    if (callPath.length > 4) {
      return this.__respond(
        $happn,
        'Failure parsing request body',
        null,
        new Error('call path cannot have more than 4 segments'),
        res,
        400
      );
    }
    this.__parseBody(req, res, $happn, (body) => {
      let authorizeAs = body?.as;
      this.__validateCredentialsGetOrigin(
        $happn,
        $origin,
        res,
        methodURI,
        authorizeAs,
        (authorizedOrigin) => {
          // will not be hit on 403
          this.__processRequest(req, res, body, callPath, $happn, authorizedOrigin);
        }
      );
    });
  } catch (e) {
    return this.__respond($happn, 'Call failed', null, e, res, 500);
  }
};

Rest.prototype.attachedToMeshEvents = false;

Rest.prototype.__buildCallMenu = function (exchangeDescription, endpoint, menu) {
  const callMenu = menu || {};
  if (!exchangeDescription.components) {
    return callMenu;
  }
  Object.keys(exchangeDescription.components)
    .filter((componentName) => !['security', 'rest'].includes(componentName))
    .forEach((componentName) => {
      const component = exchangeDescription.components[componentName];
      Object.keys(component.methods).forEach((methodName) => {
        const method = component.methods[methodName];
        let uri = '/' + componentName + '/' + methodName;
        if (endpoint) uri = '/' + endpoint + uri;
        const operation = {
          uri,
          parameters: {},
        };
        for (let param of method.parameters) {
          if (param.name === 'callback') continue;
          operation.parameters[param.name] = '{{' + param.name + '}}';
        }
        callMenu[uri] = operation;
      });
    });
  return callMenu;
};

Rest.prototype.__updateDescription = function (description, $happn) {
  const cleanedDescription = _.omit(description, ['initializing', 'setOptions', '_meta']);
  this.__exchangeDescription = cleanedDescription;
  this.__exchangeDescription.callMenu = this.__buildCallMenu(cleanedDescription);
  $happn.log.debug(`updated mesh description`);
};

Rest.prototype.start = function ($happn, callback) {
  this.__updateDescription($happn._mesh.description, $happn);
  callback();
};

Rest.prototype.initialize = function ($happn, callback) {
  this.__securityService = $happn._mesh.happn.server.services.security;
  if (!this.attachedToMeshEvents) {
    $happn._mesh.on('description-updated', (description) => {
      $happn.log.debug(`description-updated event`);
      this.__updateDescription(description, $happn);
    });
    this.attachedToMeshEvents = true;
  }
  this.__updateDescription($happn._mesh.description, $happn);
  callback();
};
